"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var lodash_1 = require("lodash");
var MapNameToId_1 = require("../generators/dot/utils/MapNameToId");
var Edge_1 = require("./Edge");
var GraphCluster_1 = require("./GraphCluster");
var GraphNode_1 = require("./GraphNode");
var MapIdToGraphNode_1 = require("./utils/MapIdToGraphNode");
/**
 * Generate a graph that represents the structure described in the tslint config.
 *
 * The graph can then be optimized before rendering (mainly to reduce the number of edges).
 *
 * note: the graph could contain circular references!
 */
var GraphGenerator = /** @class */ (function () {
    function GraphGenerator(config, filter) {
        this.config = config;
        this.filter = filter;
        // the root cluster is 0 (root is not output)
        this.containerId = 0;
        this.mapNameToId = new MapNameToId_1.MapNameToId();
        this.mapIdToNode = new MapIdToGraphNode_1.MapIdToGraphNode();
        this.root = GraphCluster_1.GraphCluster.create(null, this.containerId++, "root", "", GraphCluster_1.ClusterType.Root);
        this.mapIdToNode.add(this.root);
    }
    GraphGenerator.prototype.generateGraph = function (packageConfig) {
        var _a;
        var _this = this;
        var packageFolders = packageConfig.checkImportsBetweenPackages.packages.filter(function (pkg) {
            return _this.filter.isImportPathOkForFolder(pkg);
        });
        this.checkIfNeedsAnyPackage(packageFolders);
        var parent = this.root;
        var topLevelCluster = GraphCluster_1.GraphCluster.create(this.root, this.containerId++, "Top Level Packages", "", GraphCluster_1.ClusterType.TopLevel);
        if (!this.filter.hasWhitelist) {
            this.root.nodes.push(topLevelCluster);
        }
        parent = topLevelCluster;
        (_a = parent.nodes).push.apply(_a, this.generateNodesFromPackages(parent, packageFolders));
        packageFolders.forEach(function (pkg) {
            _this.processPackageSubFolders(_this.root, pkg);
        });
        this.processEdges(packageFolders.filter(function (pkg) { return _this.filter.canOutputSubfoldersOf(pkg); }));
        return this.root;
    };
    GraphGenerator.prototype.checkIfNeedsAnyPackage = function (packageFolders) {
        var _this = this;
        if (!this.config.dot.showImportAnyAsNodeNotEdges) {
            return;
        }
        if (packageFolders.some(function (pkg) { return pkg.allowedToImport.some(function (allowed) { return allowed === "*"; }); })) {
            packageFolders.push(this.createAnyPackage());
        }
        // also for sub-folders:
        packageFolders.forEach(function (pkg) {
            if (pkg.subFolders.some(function (sub) { return sub.allowedToImport.some(function (allowed) { return allowed === "*"; }); })) {
                pkg.subFolders.push(_this.createAnySubPackage());
            }
        });
    };
    GraphGenerator.prototype.generateNodesFromPackages = function (cluster, packageFolders) {
        var _this = this;
        if (!this.config.dot.clusterFromTslintJson) {
            return packageFolders.map(function (pkg) { return _this.generateNodeFromPackage(cluster, pkg); });
        }
        var defaultCluster = "(defaultCluster)";
        var packagesByCluster = lodash_1.groupBy(packageFolders, function (pkg) { return pkg.diagramCluster || defaultCluster; });
        var nodes = [];
        Object.keys(packagesByCluster).forEach(function (key) {
            var _a;
            var pkgs = packagesByCluster[key];
            var nodesThisCluster = pkgs.map(function (pkg) { return _this.generateNodeFromPackage(cluster, pkg); });
            if (key === defaultCluster) {
                nodes.push.apply(nodes, nodesThisCluster);
            }
            if (key !== defaultCluster) {
                var subCluster_1 = GraphCluster_1.GraphCluster.create(cluster, _this.containerId++, "", "", GraphCluster_1.ClusterType.DiagramCluster);
                _this.mapIdToNode.add(subCluster_1);
                nodes.push(subCluster_1);
                (_a = subCluster_1.nodes).push.apply(_a, nodesThisCluster);
                nodesThisCluster.forEach(function (n) { return (n.parent = subCluster_1); });
            }
        });
        return nodes;
    };
    GraphGenerator.prototype.generateNodeFromPackage = function (cluster, pkg) {
        var packageName = pkg.importPath;
        var node = this.generateNode(cluster, packageName, pkg.description, this.getNodeType(pkg));
        return node;
    };
    GraphGenerator.prototype.getNodeType = function (pkg) {
        if (pkg.isExternal) {
            return GraphNode_1.NodeType.External;
        }
        if (pkg.importPath === this.getAnyPackageId()) {
            return GraphNode_1.NodeType.Any;
        }
        return GraphNode_1.NodeType.Normal;
    };
    GraphGenerator.prototype.getNodeTypeForSubFolder = function (pkg) {
        if (pkg.importPath === this.getAnyPackageId()) {
            return GraphNode_1.NodeType.Any;
        }
        return GraphNode_1.NodeType.Normal;
    };
    /**
     * process the sub-folders of a package as a separate graph cluster.
     *
     * Else, the graph is too hard to read.
     */
    GraphGenerator.prototype.processPackageSubFolders = function (parent, pkg) {
        var _a;
        var _this = this;
        var isContainer = pkg.subFolders.length > 0;
        if (!isContainer) {
            return;
        }
        if (!this.filter.canOutputSubfoldersOf(pkg)) {
            return;
        }
        var cluster = GraphCluster_1.GraphCluster.create(parent, this.containerId++, pkg.importPath, pkg.description, GraphCluster_1.ClusterType.AreaWithSubFolders);
        this.mapIdToNode.add(cluster);
        var nodes = this.generateSubFolders(cluster, pkg.importPath, pkg.subFolders.filter(function (sub) { return _this.filter.isImportPathOkForSubFolder(sub); }));
        (_a = cluster.nodes).push.apply(_a, nodes);
        this.root.nodes.push(cluster);
    };
    GraphGenerator.prototype.generateNode = function (cluster, packageName, description, nodeType, prefix) {
        if (nodeType === void 0) { nodeType = GraphNode_1.NodeType.Normal; }
        var packageIdKey = this.getPackageIdKey(packageName, prefix);
        var packageId = this.mapNameToId.getId(packageIdKey);
        var node = new GraphNode_1.GraphNode(cluster, packageId, packageName, description, nodeType);
        this.mapIdToNode.add(node);
        return node;
    };
    GraphGenerator.prototype.getPackageIdKey = function (packageName, prefix) {
        return prefix + "_" + packageName;
    };
    GraphGenerator.prototype.processEdges = function (packageFolders) {
        var _this = this;
        packageFolders.forEach(function (pkg) {
            var thisPkgId = _this.mapNameToId.getIdOrThrow(_this.getPackageIdKey(pkg.importPath));
            var allowedPackages = pkg.allowedToImport;
            if (allowedPackages.some(function (allowed) { return allowed === "*"; })) {
                if (_this.config.dot.showImportAnyAsNodeNotEdges) {
                    allowedPackages = [_this.getAnyPackageId()];
                }
                else {
                    allowedPackages = packageFolders
                        // disallow import from self
                        .filter(function (allowed) { return allowed.importPath !== pkg.importPath; })
                        .map(function (allowed) { return allowed.importPath; });
                }
            }
            _this.processEdgesForPackageNames(thisPkgId, allowedPackages);
            // TODO review - kind of duplicate code - could have interface across PackageFolder, PackageSubFolder ?
            pkg.subFolders
                .filter(function (sub) { return _this.filter.isImportPathOkForSubFolder(sub); })
                .forEach(function (subFolder) {
                var thisSubFolderId = _this.mapNameToId.getId(_this.getPackageIdKey(subFolder.importPath, pkg.importPath));
                var allowedSubFolders = subFolder.allowedToImport;
                if (allowedSubFolders.some(function (allowed) { return allowed === "*"; })) {
                    if (_this.config.dot.showImportAnyAsNodeNotEdges) {
                        allowedSubFolders = [_this.getAnyPackageId()];
                    }
                    else {
                        allowedSubFolders = pkg.subFolders
                            .filter(
                        // disallow import from self
                        function (otherSubFolder) {
                            return otherSubFolder.importPath !== subFolder.importPath;
                        })
                            .map(function (sub) { return sub.importPath; });
                    }
                }
                _this.processEdgesForPackageNames(thisSubFolderId, allowedSubFolders, pkg.importPath);
            });
        });
    };
    GraphGenerator.prototype.getAnyPackageId = function () {
        return "*";
    };
    GraphGenerator.prototype.createAnyPackage = function () {
        return {
            allowedToImport: [],
            description: "(can import any packages)",
            importPath: this.getAnyPackageId(),
            isExternal: false,
            subFolders: []
        };
    };
    GraphGenerator.prototype.createAnySubPackage = function () {
        return {
            allowedToImport: [],
            description: "(can import any folders)",
            importPath: this.getAnyPackageId()
        };
    };
    GraphGenerator.prototype.processEdgesForPackageNames = function (thisPkgId, allowedPackages, packageIdPrefix) {
        var _this = this;
        allowedPackages
            .filter(function (pkg) { return _this.filter.isImportPathOkForEdges(pkg); })
            .forEach(function (allowedPkg) {
            var allowedPkgId = _this.mapNameToId.getIdOrThrow(_this.getPackageIdKey(allowedPkg, packageIdPrefix));
            var origin = _this.mapIdToNode.getByIdOrThrow(thisPkgId);
            var destination = _this.mapIdToNode.getByIdOrThrow(allowedPkgId);
            Edge_1.Edge.create(origin, destination);
        });
    };
    GraphGenerator.prototype.generateSubFolders = function (cluster, packageName, subFolders) {
        var _this = this;
        if (subFolders.length === 0) {
            return [];
        }
        return subFolders.map(function (folder) {
            return _this.generateNode(cluster, folder.importPath, folder.description, _this.getNodeTypeForSubFolder(folder), packageName);
        });
    };
    return GraphGenerator;
}());
exports.GraphGenerator = GraphGenerator;
//# sourceMappingURL=GraphGenerator.js.map