"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var Edge_1 = require("../Edge");
var GraphCluster_1 = require("../GraphCluster");
var GraphVisitor_1 = require("./GraphVisitor");
/**
 * Optimize the given graph, to reduce the number of edges so that it is more readable.
 *
 * Algorithm:
 *
 * [done] 1a. all nodes with same set of incoming edges -> place in an 'optimization' cluster, and replace those edges with 1 edge to the cluster
 * 1b. edges with same origin, and destinations are all in same *optimization* cluster -> replace with 1 edge to that cluster
 * [done] 1c. edges with same destination, and origins are all in same *optimization* cluster -> replace with 1 edge from that cluster
 * (future) could detect nodes with *mostly* same incoming edges -> place in cluster, and replace *some* edges
 * (alt option) nodes in same cluster as 'records' (just a rendering option?)
 */
var GraphOptimizer = /** @class */ (function () {
    function GraphOptimizer() {
        this.containerId = 1;
    }
    GraphOptimizer.prototype.optimize = function (root) {
        this.optimizeNodesWithSameIncomings(root);
        // TODO xxx optimization 1b
        this.optimizeEdgesWithOriginsInSameCluster(root);
    };
    GraphOptimizer.prototype.optimizeNodesWithSameIncomings = function (root) {
        // 1a. all nodes with same set of incoming edges -> place in cluster, and replace those edges with 1 edge to the cluster
        var _this = this;
        var _a = this.populateMaps(root), mapNodeIdToNode = _a.mapNodeIdToNode, mapNodeIdToOrigins = _a.mapNodeIdToOrigins;
        // build map from 'same origin IDs' to nodes
        var mapOriginIdSetToNodeIds = new Map();
        mapNodeIdToOrigins.forEach(function (value, nodeId) {
            var originIds = value.join(",");
            if (!mapOriginIdSetToNodeIds.get(originIds)) {
                mapOriginIdSetToNodeIds.set(originIds, []);
            }
            mapOriginIdSetToNodeIds.get(originIds).push(nodeId);
        });
        mapOriginIdSetToNodeIds.forEach(function (value, originIds) {
            var nodeIds = mapOriginIdSetToNodeIds.get(originIds);
            if (nodeIds.length > 1) {
                // We have a set of nodes that can be clustered
                root.nodes
                    .filter(function (node) { return node instanceof GraphCluster_1.GraphCluster; })
                    .forEach(function (node) {
                    var cluster = node;
                    _this.optimizeNodesWithSameIncomingsForCluster(cluster, nodeIds, mapNodeIdToNode);
                });
            }
        });
    };
    GraphOptimizer.prototype.populateMaps = function (root) {
        var visitor = new GraphVisitor_1.GraphVisitor(root);
        var mapNodeIdToOrigins = new Map();
        var mapNodeIdToNode = new Map();
        var visit = function (node) {
            mapNodeIdToNode.set(node.id, node);
            node.incomingEdges.forEach(function (edge) {
                if (!mapNodeIdToOrigins.get(node.id)) {
                    mapNodeIdToOrigins.set(node.id, []);
                }
                var list = mapNodeIdToOrigins.get(node.id);
                list.push(edge.origin.id);
            });
        };
        visitor.visitNodesAndClusters(visit);
        return {
            mapNodeIdToNode: mapNodeIdToNode,
            mapNodeIdToOrigins: mapNodeIdToOrigins
        };
    };
    GraphOptimizer.prototype.optimizeNodesWithSameIncomingsForCluster = function (parent, nodeIds, mapNodeIdToNode) {
        var _a;
        var matchingNodeIdsInParent = nodeIds.filter(function (node) {
            return parent.nodes.find(function (n) { return n.id === node; });
        });
        var nodes = matchingNodeIdsInParent.map(function (id) {
            var node = mapNodeIdToNode.get(id);
            if (!node) {
                throw new Error("cannot find node from map for id " + id);
            }
            return node;
        });
        // remove the nodes from the parent:
        nodes.forEach(function (node) { return parent.removeNode(node); });
        // remove the old edges, and build list of origin nodes:
        var originNodes = [];
        nodes.forEach(function (node) {
            // copy list of edges, since we are deleting from it:
            var incomingEdges = __spreadArrays(node.incomingEdges);
            incomingEdges.forEach(function (edge) {
                edge.remove();
                originNodes.push(edge.origin);
            });
        });
        // Add the new cluster:
        var clusterId = "CO1_" + this.containerId++;
        var newCluser = new GraphCluster_1.GraphCluster(parent, clusterId, "");
        newCluser.clusterType = GraphCluster_1.ClusterType.FromOptimization;
        (_a = newCluser.nodes).push.apply(_a, nodes);
        nodes.forEach(function (node) { return (node.parent = newCluser); });
        parent.nodes.push(newCluser);
        // add the new edges:
        originNodes.forEach(function (origin) {
            Edge_1.Edge.create(origin, newCluser);
        });
    };
    GraphOptimizer.prototype.optimizeEdgesWithOriginsInSameCluster = function (root) {
        // 1c. edges with same destination, and origins are all in same *optimization* cluster -> replace with 1 edge from that cluster
        // For a node, find all the edges where the origin is in the same cluster
        var mapNodeIdToNode = this.populateMaps(root).mapNodeIdToNode;
        mapNodeIdToNode.forEach(function (node) {
            if (node.incomingEdges.length < 2 || !node.parent) {
                return;
            }
            var mapEdgesByIncomingClusterId = new Map();
            node.incomingEdges.forEach(function (edge) {
                if (!edge.origin.parent) {
                    return;
                }
                if (node.parent.id === edge.origin.parent.id) {
                    // node and edge origin are in same cluster
                    return;
                }
                var incomingClusterId = edge.origin.parent.id;
                if (!mapEdgesByIncomingClusterId.has(incomingClusterId)) {
                    mapEdgesByIncomingClusterId.set(incomingClusterId, []);
                }
                mapEdgesByIncomingClusterId.get(incomingClusterId).push(edge);
            });
            // For cluster IDs that have multiple edges, replace those edges with 1:
            mapEdgesByIncomingClusterId.forEach(function (edges, clusterId) {
                if (edges.length < 2) {
                    return;
                }
                var incomingCluster = mapNodeIdToNode.get(clusterId);
                if (!incomingCluster) {
                    throw new Error("cannot find cluster id " + clusterId);
                }
                if (!(incomingCluster instanceof GraphCluster_1.GraphCluster)) {
                    throw new Error("incoming cluster is not a cluster! - id " + clusterId);
                }
                // only create an edge from that cluster, if ALL the nodes in the cluster have outgoing to this node:
                var allClusterNodesHaveOutgoingsToThisNode = incomingCluster.nodes.every(function (n) { return n.outgoingEdges.some(function (e) { return e.destination.id === node.id; }); });
                if (!allClusterNodesHaveOutgoingsToThisNode) {
                    return;
                }
                // remove the edges
                edges.forEach(function (edge) { return edge.remove(); });
                // add 1 edge from the cluster, to the node
                Edge_1.Edge.create(incomingCluster, node);
            });
        });
    };
    return GraphOptimizer;
}());
exports.GraphOptimizer = GraphOptimizer;
//# sourceMappingURL=GraphOptimizer.js.map